MODULE RK;

IMPORT MathL;

CONST minScale = 0.125; maxScale = 8.0; attempts = 12;

TYPE
    State* = POINTER TO ARRAY OF FLOAT64;
    (*DynamicalSystem* = POINTER TO DynamicalSystemVariables;*)
    DynamicalSystem* = RECORD
    (** May be extended to include external parameters *)
        state*: State;
    END;

PROCEDURE (VAR ds: DynamicalSystem) ComputeRHS*(result: State);
(** Must be redefined by the target application, for now does nothing *)
BEGIN
END ComputeRHS;

PROCEDURE Copy(input: State): State;
VAR i: SIZE; result: State;
BEGIN
	NEW(result,LEN(input^));
    FOR i := 0 TO LEN(input^)-1 DO
        result^[i] := input^[i]
    END;
    RETURN result;
END Copy;

PROCEDURE (VAR ds: DynamicalSystem) GenerateK(k1,k3,k4,k5,k6: State; h: FLOAT64; generateK6: BOOLEAN);
CONST   C31 = 3/32; C32 = 3*C31; C41 = 1932/2197; C42 = -7200/2197; C43 = 7296/2197;
        C51 = 439/216; C52 = -8.0; C53 = 3680/513; C54 = -845/4104;
        C61 = -8/27; C62 = 2.0; C63 = -3544/2565; C64 = 1859/4104; C65 = -11/40;
VAR k2, y0: State; i, L: SIZE;
BEGIN
    L := LEN(ds.state^);
    NEW(y0,L); y0 := Copy(ds.state);
    NEW(k2,L);
    ds.ComputeRHS(k1);
    FOR i := 0 TO L-1 DO
        ds.state[i] := y0[i] + h/4*k1[i]
    END;
    ds.ComputeRHS(k2);
    FOR i := 0 TO L-1 DO
        ds.state[i] := y0[i] + h*(C32*k2[i] + C31*k1[i])
    END;
    ds.ComputeRHS(k3);
    FOR i := 0 TO L-1 DO
        ds.state[i] := y0[i] + h*(C43*k3[i] + C42*k2[i] + C41*k1[i])
    END;
    ds.ComputeRHS(k4);
    FOR i := 0 TO L-1 DO
        ds.state[i] := y0[i] + h*(C54*k4[i] + C53*k3[i] + C52*k2[i] + C51*k1[i])
    END;
    ds.ComputeRHS(k5);
    IF generateK6 THEN
        FOR i := 0 TO L-1 DO
            ds.state[i] := y0[i] + h*(C65*k5[i] + C64*k4[i] + C63*k3[i] + C62*k2[i] + C61*k1[i])
        END;
        ds.ComputeRHS(k6);
    END;
    ds.state := Copy(y0); (* restoring the ds initial state *)
END GenerateK;

PROCEDURE (VAR ds: DynamicalSystem) UpdateRK4K(k1,k3,k4,k5: State; step: FLOAT64);
CONST C1 = 25/216; C3 = 1408/2565; C4 = 2197/4104; C5 = -1/5;
VAR i: SIZE;
BEGIN
    FOR i := 0 TO LEN(ds.state^)-1 DO
        ds.state[i] := ds.state[i] + step*(C5*k5[i] + C4*k4[i] + C3*k3[i] + C1*k1[i])
    END;
END UpdateRK4K;

PROCEDURE NormSquare(vector: State): FLOAT64;
VAR i: SIZE; normSquare, component: FLOAT64;
BEGIN
	normSquare := 0.0;
    FOR i := 0 TO LEN(vector^)-1 DO
        component := vector^[i];
        normSquare := normSquare + component*component;
    END;
    RETURN normSquare;
END NormSquare;

PROCEDURE ErrorSquare(k1,k3,k4,k5,k6: State): FLOAT64;
CONST C1 = 1/360; C3 = -128/4275; C4 = -2197/75240; C5 = 0.02; C6 = 2/55;
VAR errorVector: State; i: SIZE;
BEGIN
    NEW(errorVector,LEN(k1^));
    FOR i := 0 TO LEN(errorVector^)-1 DO
        errorVector[i] := C6*k6[i] + C5*k5[i] + C4*k4[i] + C3*k3[i] + C1*k1[i]
    END;
    RETURN NormSquare(errorVector);
END ErrorSquare;

PROCEDURE (VAR ds: DynamicalSystem) UpdateRK4*(step: FLOAT64);
(** An application of the Runge-Kutta 4 method with time-step 'step'.
    The array of state and of the right hand side functions must
    be defined in the calling application
*)
VAR k1, k3, k4, k5, k6: State; j: SIZE;
BEGIN
    j := LEN(ds.state^);
    NEW(k1,j); NEW(k3,j); NEW(k4,j); NEW(k5,j); k6 := NIL;
    ds.GenerateK(k1,k3,k4,k5,k6,step,FALSE);
    ds.UpdateRK4K(k1,k3,k4,k5,step);
END UpdateRK4;

PROCEDURE (VAR ds: DynamicalSystem) UpdateRKF45*(VAR step, oldStep: FLOAT64; rtol, atol: FLOAT64);
(** An application of the Runge-Kutta-Fehlberg 45 method with the initial time-step guess 'step'.
    The array of state and of the right hand side functions must be defined in the calling application.
    Internally RK4 and RK5 are performed and the result is compared. If significantly different (w.r.t.
    to the tolerance), 'step' is updated. The new 'step' estimate (for the next application) is returned
    alongside with the updated 'state'. The actually used step during the 'state' update is returned in
    'oldStep'. 'rtol' is the relative error tolerance, 'atol' is the absolute error tolerance
*)
VAR k1, k3, k4, k5, k6, y: State; j, i: SIZE; scale: FLOAT64;
	error, value: FLOAT64; (* square of the error and the state norm *)
BEGIN
    j := LEN(ds.state^); NEW(y,j);
    NEW(k1,j); NEW(k3,j); NEW(k4,j); NEW(k5,j); NEW(k6,j);
    rtol := rtol*rtol; (* we work with squares of the norms *)
    atol := atol*atol; (* we work with squares of the norms *)
    i := 0; y := Copy(ds.state);
    REPEAT
        oldStep := step;
        ds.state := Copy(y);
        ds.GenerateK(k1,k3,k4,k5,k6,oldStep,TRUE);
        error := ErrorSquare(k1,k3,k4,k5,k6);
        ds.UpdateRK4K(k1,k3,k4,k5,oldStep);
        value := rtol*NormSquare(ds.state);
        IF value < atol THEN value := atol ELSIF value = 0.0 THEN value := rtol END;
        scale := 0.84*MathL.sqrt(MathL.sqrt(value/error));
        IF scale < minScale THEN scale := minScale ELSIF scale > maxScale THEN scale := maxScale END;
        step := scale*oldStep;
        INC(i);
    UNTIL (error < value) OR (i = attempts);
END UpdateRKF45;

END RK.
